/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: overlay_gapi.c 200 2005-01-14 22:09:26Z picard $
 *
 * BetaPlayer Overlay plugin
 * Copyright (c) 2004 Gabor Kovacs
 *
 ****************************************************************************/

#include "stdafx.h"

#ifdef _WIN32

#define WIN32_LEAN_AND_MEAN
#ifndef STRICT
#define STRICT
#endif
#include <windows.h>

/*modify*/#include "overlay_subtitle.h"

typedef struct GXDisplayProperties 
{
	int cxWidth;
	int cyHeight;
	int cbxPitch;
	int cbyPitch;
	int cBPP;
	int ffFormat;
} GXDisplayProperties;

#define SHFS_SHOWTASKBAR            0x0001
#define SHFS_HIDETASKBAR            0x0002
#define SHFS_SHOWSIPBUTTON          0x0004
#define SHFS_HIDESIPBUTTON          0x0008
#define SHFS_SHOWSTARTICON          0x0010
#define SHFS_HIDESTARTICON          0x0020

#define GX_FULLSCREEN		0x01
#define kfLandscape			0x08
#define kfPalette			0x10
#define kfDirect			0x20
#define kfDirect555			0x40
#define kfDirect565			0x80
#define kfDirect888			0x100
#define kfDirect444			0x200
#define kfDirectInverted	0x400

typedef struct gapi
{
	overlay_base p;
	int AdjustPtr;
	uint8_t* Pointer;
	uint8_t* DirectPtr;
	bool_t DRAM;
	bool_t Windows;
	bool_t Opened;
	GXDisplayProperties Info;
	HMODULE AygShell;

	int (*GXOpenDisplay)(HWND hWnd, DWORD dwFlags);
	int (*GXCloseDisplay)();
	void* (*GXBeginDraw)();
	int (*GXEndDraw)();
	GXDisplayProperties (*GXGetDisplayProperties)();
	int (*GXSetViewport)( DWORD dwTop, DWORD dwHeight, DWORD, DWORD );
	BOOL (*GXIsDisplayDRAMBuffer)();
	BOOL (WINAPI* SHFullScreen)( HWND, DWORD );

	/*modify*/
	node *s;
	planes Buffer[3];
	int BufDraw;
	int BufLast;
	int BufClear;
	bool_t BufOn;
	bool_t BufDiff;
	/*modify end*/
} gapi;

/*modify*/
#define	GAPI_SUB_BUFFER			0x300
#define	GAPI_SUB_BUFFER_DIFF	0x301
/*modify end*/

const datadef GAPIParams[] = 
{
	{ GAPI_ID,			TYPE_CLASS },
	{ GAPI_WIDTH,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ GAPI_HEIGHT,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ GAPI_PITCHX,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ GAPI_PITCHY,		TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ GAPI_BPP,			TYPE_INT, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ GAPI_FORMAT,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX | DF_HIDDEN },
	{ GAPI_DRAM,		TYPE_BOOL, DF_SETUP | DF_RDONLY | DF_HIDDEN },
	{ GAPI_POINTER,		TYPE_INT, DF_SETUP | DF_RDONLY|DF_HEX | DF_HIDDEN },
	/*modify*/
	{ GAPI_SUB_BUFFER,		TYPE_BOOL, DF_SETUP | DF_CHECKLIST },
	{ GAPI_SUB_BUFFER_DIFF,	TYPE_BOOL, DF_SETUP | DF_CHECKLIST },
	/*modify end*/
	DATADEF_END
};

/*modify*/static int BufferAlign(gapi* p)
{
	GXDisplayProperties Info = p->GXGetDisplayProperties();

	if((!p->Buffer[0][0])&&(p->Buffer[1][0])){
		p->Buffer[0][0] = p->Buffer[1][0];
		p->Buffer[1][0] = NULL;
	}

	if(p->BufDiff){
		if(!p->Buffer[0][0]) p->Buffer[0][0] = Alloc16(Info.cxWidth * Info.cyHeight * Info.cBPP >> 3);
		if(!p->Buffer[1][0]) p->Buffer[1][0] = Alloc16(Info.cxWidth * Info.cyHeight * Info.cBPP >> 3);
	} else if(p->BufOn){
		if(!p->Buffer[0][0]) p->Buffer[0][0] = Alloc16(Info.cxWidth * Info.cyHeight * Info.cBPP >> 3);
		if(p->Buffer[1][0]) { Free16(p->Buffer[1][0]);p->Buffer[1][0] = NULL; }
	} else {
		if(p->Buffer[0][0]) { Free16(p->Buffer[0][0]);p->Buffer[0][0] = NULL; }
		if(p->Buffer[1][0]) { Free16(p->Buffer[1][0]);p->Buffer[1][0] = NULL; }
	}

	if((!p->Buffer[0][0])&&(p->Buffer[1][0])){
		p->Buffer[0][0] = p->Buffer[1][0];
		p->Buffer[1][0] = NULL;
	}

	p->BufDraw = 0;
	p->BufLast = 2;
	p->BufClear = 1;

	return ERR_NONE;
}

static int Get(gapi* p,int No,void* Data,int Size)
{
	int Result = OvlGet(&p->p,No,Data,Size);
	switch (No)
	{
	case GAPI_WIDTH: GETVALUE(p->Info.cxWidth,int); break;
	case GAPI_HEIGHT: GETVALUE(p->Info.cyHeight,int); break;
	case GAPI_PITCHX: GETVALUE(p->Info.cbxPitch,int); break;
	case GAPI_PITCHY: GETVALUE(p->Info.cbyPitch,int); break;
	case GAPI_BPP: GETVALUE(p->Info.cBPP,int); break;
	case GAPI_FORMAT: GETVALUE(p->Info.ffFormat,int); break;
	case GAPI_DRAM: GETVALUE(p->DRAM,bool_t); break;
	case GAPI_POINTER: GETVALUE((int)(p->Pointer + p->AdjustPtr),int); break;
	/*modify*/
	case GAPI_SUB_BUFFER: GETVALUE(p->BufOn, bool_t); break;
	case GAPI_SUB_BUFFER_DIFF: GETVALUE(p->BufDiff,bool_t); break;
	/*modify end*/
	}
	return Result;
}

/*modify*/static int Set(gapi* p,int No,const void* Data,int Size)
{
	int Result;
	switch (No)
	{
	case GAPI_SUB_BUFFER: SETVALUE(p->BufOn,bool_t,Result = BufferAlign(p)); break;
	case GAPI_SUB_BUFFER_DIFF: SETVALUE(p->BufDiff,bool_t,Result = BufferAlign(p)); break;
	default: Result = OvlSet(&p->p, No, Data, Size); break;
	}
	return Result;
}

static void LoadDLL(gapi* p, bool_t Windows )
{
	tchar_t Name[MAXPATH];
	HMODULE Module;
	p->Windows = Windows;

	if (Windows)
	{
		tcscpy(Name,GetSystemPath());
		tcscat(Name,T("\\gx.dll"));
	}
	else
		tcscpy(Name,T("gx.dll"));

	Module = LoadLibrary(Name);
	if (Module)
	{
		if (p->p.Module)
			FreeLibrary(p->p.Module);

		p->p.Module = Module;
		GetProc(&p->p.Module,&p->GXOpenDisplay,T("?GXOpenDisplay@@YAHPAUHWND__@@K@Z"),0);
		GetProc(&p->p.Module,&p->GXCloseDisplay,T("?GXCloseDisplay@@YAHXZ"),0);
		GetProc(&p->p.Module,&p->GXBeginDraw,T("?GXBeginDraw@@YAPAXXZ"),0);
		GetProc(&p->p.Module,&p->GXEndDraw,T("?GXEndDraw@@YAHXZ"),0);
		GetProc(&p->p.Module,&p->GXGetDisplayProperties,T("?GXGetDisplayProperties@@YA?AUGXDisplayProperties@@XZ"),0);
		GetProc(&p->p.Module,&p->GXSetViewport,T("?GXSetViewport@@YAHKKKK@Z"),1);
		GetProc(&p->p.Module,&p->GXIsDisplayDRAMBuffer,T("?GXIsDisplayDRAMBuffer@@YAHXZ"),1);
	}
}

static int Init(gapi* p)
{
	int Caps = QueryPlatform(PLATFORM_CAPS);
	GXDisplayProperties Info;
	video GDI;

	if (!p->Opened)
	{
		NodeError((node*)p,GAPI_ID,GAPI_OPEN_ERROR);
		return ERR_NOT_SUPPORTED; // don't show device error
	}

	Info = p->Info;

	memset(&p->p.Output,0,sizeof(p->p.Output));
	p->p.Output.Direction = 0;
	p->p.Output.Aspect = ASPECT_ONE;
	p->Pointer = p->DirectPtr;

	if (Info.ffFormat & kfPalette)
	{
		QueryDesktop(&GDI);

		p->p.Output.Pixel.Flags = PF_PALETTE;
		p->p.Output.Pixel.BitCount = Info.cBPP;
		p->p.Output.Pixel.Palette = GDI.Pixel.Palette;
	}
	else
	if (Info.ffFormat & kfDirect444)
		DefaultRGB(&p->p.Output.Pixel,Info.cBPP,4,4,4,0,0,0);
	else
	if (Info.ffFormat & kfDirect565)
	{
		if (Caps & CAPS_ONLY12BITRGB)
			DefaultRGB(&p->p.Output.Pixel,Info.cBPP,4,4,4,1,2,1);
		else
			DefaultRGB(&p->p.Output.Pixel,Info.cBPP,5,6,5,0,0,0);
	}
	else
	if (Info.ffFormat & kfDirect555)
		DefaultRGB(&p->p.Output.Pixel,Info.cBPP,5,5,5,0,0,0);
	else
	if (Info.ffFormat & kfDirect888)
		DefaultRGB(&p->p.Output.Pixel,Info.cBPP,8,8,8,0,0,0);

	if (Info.ffFormat & kfDirectInverted)
		p->p.Output.Pixel.Flags |= PF_INVERTED;

	// get signed x/y pitches in bits
	Info.cbxPitch <<= 3;
	Info.cbyPitch <<= 3;

	// one pitch is probably zero when BPP<8
	if (!Info.cbxPitch) 
		if (Info.cbyPitch < 0)
			Info.cbxPitch = -Info.cBPP;
		else
			Info.cbxPitch = Info.cBPP;
	else
	if (!Info.cbyPitch)
		if (Info.cbxPitch < 0)
			Info.cbyPitch = Info.cBPP;
		else
			Info.cbyPitch = -Info.cBPP;

	// one of the pitches absolute value must be BPP
	if (abs(Info.cbyPitch) < abs(Info.cbxPitch))
	{
		if (abs(Info.cbxPitch) < Info.cyHeight*Info.cBPP &&
			abs(Info.cbxPitch) >= Info.cxWidth*Info.cBPP) //swapped gapi resolution
			SwapInt(&Info.cxWidth,&Info.cyHeight);

		Info.ffFormat |= kfLandscape;
		Info.cbyPitch = Info.cbyPitch<0 ? -Info.cBPP : Info.cBPP;
	}
	else
	{
		if (abs(Info.cbyPitch) < Info.cxWidth*Info.cBPP &&
			abs(Info.cbyPitch) >= Info.cyHeight*Info.cBPP) //swapped gapi resolution
			SwapInt(&Info.cxWidth,&Info.cyHeight);

		Info.ffFormat &= ~kfLandscape;
		Info.cbxPitch = Info.cbxPitch<0 ? -Info.cBPP : Info.cBPP;
	}

	p->p.Output.Width = Info.cxWidth;
	p->p.Output.Height = Info.cyHeight;

	// we need the physical start of the framebuffer
	p->AdjustPtr = 0;
	if (Info.cbxPitch<0) 
		p->AdjustPtr += (Info.cbxPitch * (Info.cxWidth-1)) >> 3;
	if (Info.cbyPitch<0) 
		p->AdjustPtr += (Info.cbyPitch * (Info.cyHeight-1)) >> 3;

	if (Info.ffFormat & kfLandscape)
	{
		p->p.Output.Direction |= DIR_SWAPXY;
		p->p.Output.Pitch = abs(Info.cbxPitch) >> 3;
		SwapInt(&p->p.Output.Width,&p->p.Output.Height);

		if (Info.cbxPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
		if (Info.cbyPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
	}
	else
	{
		p->p.Output.Pitch = abs(Info.cbyPitch) >> 3;

		if (Info.cbxPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORLEFTRIGHT;
		if (Info.cbyPitch<0) 
			p->p.Output.Direction |= DIR_MIRRORUPDOWN;
	}

	if (!p->DRAM)
	{
		p->p.SetFX = BLITFX_AVOIDTEARING;

		if ((Info.cxWidth > Info.cyHeight) && (Caps & CAPS_PORTRAIT))
			p->p.SetFX |= BLITFX_VMEMROTATED;

		if ((Info.cyHeight > Info.cxWidth) && (Caps & CAPS_LANDSCAPE))
			p->p.SetFX |= BLITFX_VMEMROTATED;
	}
	else
		p->p.ClearFX = BLITFX_ONLYDIFF;

	// wm2003se gapi emulation?
	if (QueryPlatform(PLATFORM_VER) >= 421)
	{
		if (p->p.Output.Width == 240 && p->p.Output.Height == 320 && p->p.Output.Pitch == 640)
			p->p.Output.Pitch = 480;
	}

	if (Info.cxWidth == 240 && Info.cyHeight == 320)
		OvlAdjustOrientation(&p->p,0);

	/*modify*/p->s = NodeFind(SUBT_ID);
	return ERR_NONE;
}

static int UpdateWnd(gapi* p)
{
	if (p->p.Wnd)
	{
		if (!p->Opened)
		{
			HWND WndTask;

			if (!p->p.Module || !p->GXOpenDisplay(p->p.Wnd,GX_FULLSCREEN))
			{
				// try gx.dll in \windows directory (HPC with .NET)
				LoadDLL(p,1);
				if (!p->p.Module || !p->GXOpenDisplay(p->p.Wnd,GX_FULLSCREEN))
					return ERR_NOT_SUPPORTED;
			}

			// just in case GXOpenDisplay hides these things
			if (p->SHFullScreen)
			{
				p->SHFullScreen(p->p.Wnd,SHFS_SHOWSTARTICON);
				p->SHFullScreen(p->p.Wnd,SHFS_SHOWTASKBAR);
			}

			WndTask = FindWindow(T("HHTaskbar"),NULL);
			if (WndTask)
				ShowWindow(WndTask,SW_SHOWNA);

			p->Info = p->GXGetDisplayProperties();
			if (!p->Info.cxWidth || !p->Info.cyHeight)
			{
				// this shouldn't happen
				p->GXCloseDisplay();
				return ERR_DEVICE_ERROR;
			}

			p->Opened = 1;

			p->DRAM = 0;
			if (p->GXIsDisplayDRAMBuffer)
				p->DRAM = p->GXIsDisplayDRAMBuffer() != 0;

			// detect fake DRAM
			if (p->DRAM && !(p->Info.ffFormat & kfLandscape) && QueryPlatform(PLATFORM_MODEL)==MODEL_SPV_C500_ORIGROM)
			{
				p->GXSetViewport(0,0,0,0);
				p->DirectPtr = p->GXBeginDraw();
			}
			if (QueryPlatform(PLATFORM_VER) >= 421)
				IsOrientationChanged(); // force requery of orientation (GXOpenDisplay changes to portrait)
		}
	}
	else
	{
		if (p->Opened && !p->DirectPtr)
		{
			p->GXCloseDisplay();
			p->Opened = 0;
		}
	}

	return ERR_NONE;
}

/*modify*/int Update( gapi* p )
{
	if(p->s){
		RedrawSubtitle(p->s);
		p->BufDraw = 0;
		p->BufLast = 2;
		p->BufClear = 1;
	}
	return OvlUpdateAlign(&(p->p));
}

static void Done(gapi* p)
{
}

static int Reset(gapi* p)
{
	Init(p);
	Done(p);
	return ERR_NONE;
}

static int Lock(gapi* p, planes Planes, bool_t OnlyAligned )
{
	if (!p->Opened)
		return ERR_INVALID_PARAM;

	if (p->DirectPtr)
		Planes[0] = p->DirectPtr + p->AdjustPtr;
	else
	{
		if (p->DRAM && p->GXSetViewport)
		{
			int Top,Height;
			if (OnlyAligned)
			{
				Top = p->p.GUIAlignedRect.y;
				Height = p->p.GUIAlignedRect.Height;
			}
			else
			{
				Top = p->p.Viewport.y;
				Height = p->p.Viewport.Height;
			}

			if (p->p.Output.Pixel.Flags & PF_PIXELDOUBLE)
			{
				Top >>= 1;
				Height >>= 1;
			}

			p->GXSetViewport(Top,Height,0,0); // coords in windows space
		}

		p->Pointer = (uint8_t*)p->GXBeginDraw();
		if (!p->Pointer)
			return ERR_NOT_SUPPORTED;

		Planes[0] = p->Pointer + p->AdjustPtr;
	}

	return ERR_NONE;
}

static int Unlock(gapi* p )
{
	if (!p->DirectPtr)
		p->GXEndDraw();
	return ERR_NONE;
}

/*modify*/void PlaneCopy(gapi* p, planes Dst, planes Src, planes Last){
	int Width, Height, DstX, DstY;
	int x,y;
	int BPP;

	Width = p->p.DstAlignedRect.Width;
	Height = p->p.DstAlignedRect.Height;
	DstX = p->p.DstAlignedRect.x;
	DstY = p->p.DstAlignedRect.y;
	BPP = p->Info.cBPP;

	/*if(p->BufClear){
		int Pitch = Pack->Dst.Pitch;
		char *SrcPtr = (char*)Src[0] + (DstY * Pitch) + (DstX * Pack->Code[0].DstBPP >> 3);
		char *DstPtr = (char*)Dst[0] + (DstY * Pitch);
		int Line = Width * Pack->Code[0].DstBPP >> 3;
		int Gap = Pitch - (Width * Pack->Code[0].DstBPP >> 3);
		memset(DstPtr, 0, DstX * Pack->Code[0].DstBPP >> 3);
		DstPtr += DstX * Pack->Code[0].DstBPP >> 3;
		for(y=0;y<Height-1;y++){
			memcpy(DstPtr, SrcPtr, Line);
			memset(DstPtr + Line, 0, Gap);
			DstPtr += Pitch;
			SrcPtr += Pitch;
		}
		memcpy(DstPtr, SrcPtr, Line);
		memset(DstPtr + Line, 0, Gap - (DstX * Pack->Code[0].DstBPP >> 3));
		p->BufClear = 0;
	} else */
	if((Last == NULL)||(Last[0] == NULL)){
		int Pitch = p->p.Output.Pitch;
		char *SrcPtr = (char*)Src[0] + (DstY * Pitch) + (DstX * BPP >> 3);
		char *DstPtr = (char*)Dst[0] + (DstY * Pitch) + (DstX * BPP >> 3);
		int Line = Width * BPP >> 3;

		for(y=0;y<Height;y++){
			memcpy(DstPtr, SrcPtr, Line);
			DstPtr += Pitch;
			SrcPtr += Pitch;
		}
	} else {
		// fix me!! (need optimize)
		switch(BPP){
			case 8:{
				int Pitch = p->p.Output.Pitch;
				char *SrcPtr = (char*)Src[0] + DstY * Pitch + DstX;
				char *DstPtr = (char*)Dst[0] + DstY * Pitch + DstX;
				char *LastPtr = (char*)Last[0] + DstY * Pitch + DstX;
				int Gap = Pitch - Width;
				for(y=Height;y;y--){
					for(x=Width;x;x--, SrcPtr++, DstPtr++, LastPtr++){
						if(*SrcPtr != *LastPtr){
							*DstPtr = *SrcPtr;
						}
					}
					SrcPtr += Gap;
					DstPtr += Gap;
					LastPtr += Gap;
				}
				break;
			}
			case 16:{
				int Pitch = p->p.Output.Pitch >> 1;
				short *SrcPtr = (short*)Src[0] + DstY * Pitch + DstX;
				short *DstPtr = (short*)Dst[0] + DstY * Pitch + DstX;
				short *LastPtr = (short*)Last[0] + DstY * Pitch + DstX;
				for(y=Height;y;y--){
					int *SrcLine = (int*)SrcPtr;
					int *DstLine = (int*)DstPtr;
					int *LastLine = (int*)LastPtr;
					for(x=Width>>1;x;x--, SrcLine++, DstLine++, LastLine++){
						if(*SrcLine != *LastLine)
							*DstLine = *SrcLine;
					}
					if((Width & 1)&&(*(short*)SrcLine != *(short*)LastLine))
						*(short*)DstLine = *(short*)SrcLine;
					SrcPtr += Pitch;
					DstPtr += Pitch;
					LastPtr += Pitch;
				}
				break;
			}
		};
	}

	if(p->Buffer[1][0]) {
		p->BufLast = p->BufDraw;
		p->BufDraw = 1 - p->BufLast;
	}
	return;
}

/*modify*/static int Blit(gapi* p, const planes Data, const planes DataLast )
{
	planes Planes;
	int Result;

	Result = Lock(p, Planes, 1);

	if (Result==ERR_NONE)
	{
		if(p->s){
			int Redraw;
			GetSubtitlePos(p->s, p->p.LastTime);
			Redraw = DrawSubtitle(p->s, &p->p, SUBTITLE_GAPI, -1);
			if(p->Buffer[1][0]){
				// Double buffer and diffrent change
				if(Redraw == -1){
					BlitImage(p->p.Soft, Planes, Data, DataLast, -1, -1);
					p->BufDraw = 0;
					p->BufLast = 2;
					p->BufClear = 1;
				} else {
					if(Redraw == 1)
						BlitImage(p->p.Soft, p->Buffer[p->BufDraw], Data, NULL, -1, -1);
					else
						BlitImage(p->p.Soft, p->Buffer[p->BufDraw], Data, DataLast, -1, -1);
					BlitSubtitle(p->s, p->Buffer[p->BufDraw][0]);
					if((DataLast == NULL)||(DataLast[0] == NULL))
						PlaneCopy(p, Planes, p->Buffer[p->BufDraw], NULL);
					else
						PlaneCopy(p, Planes, p->Buffer[p->BufDraw], p->Buffer[p->BufLast]);
				}
			} else if(Redraw == -1){
				BlitImage(p->p.Soft, Planes, Data, DataLast, -1, -1);
			} else if(p->Buffer[0][0]) {
				// Single buffer
				if(Redraw == 1)
					BlitImage(p->p.Soft, p->Buffer[0], Data, NULL, -1, -1);
				else
					BlitImage(p->p.Soft, p->Buffer[0], Data, DataLast, -1, -1);
				BlitSubtitle(p->s, p->Buffer[0][0]);
				PlaneCopy(p, Planes, p->Buffer[0], NULL);
			} else {
				// Not buffering
				if(Redraw == 1)
					BlitImage(p->p.Soft, Planes, Data, NULL, -1, -1);
				else
					BlitImage(p->p.Soft, Planes, Data, DataLast, -1, -1);
				BlitSubtitle(p->s, Planes[0]);
			};
		} else {
			BlitImage(p->p.Soft, Planes, Data, DataLast, -1, -1);
		}

		Unlock(p);
	}

	return Result;
}

static gapi GAPI = 
{
	NULL,
	Get,
	/*modift*/Set,
	Init,
	Done,
	/*modify*/Blit, //blitimage
	/*modify*/Update, //updatealign
	NULL, //updateshow
	NULL, //queryorientation
	Reset,
	UpdateWnd,
	Lock,
	Unlock,
};

void GAPI_Init() 
{ 
	int Pri = PRI_DEFAULT+50;
	gapi* p = &GAPI;

	if (NodeFind(RAW_ID))
		return; // disable when Raw Framebuffer is ready available

	p->AygShell = LoadLibrary(T("aygshell.dll"));
	GetProc(&p->AygShell,&p->SHFullScreen,T("SHFullScreen"),1);

	p->p.Params = GAPIParams;
	LoadDLL(p,0);

	if (QueryPlatform(PLATFORM_MODEL)==MODEL_SPV_C500_ORIGROM)
		Pri -= 55; // use GDI by default until problem is found with C500

	/*modify*/
	p->BufOn = 0;
	p->BufDiff = 0;
	p->Buffer[0][0] = NULL;
	p->Buffer[1][0] = NULL;
	p->Buffer[2][0] = NULL;

	p->s = NULL;
	/*modify end*/

	OvlInit(&GAPI.p,GAPI_ID,1,Pri);
}

void GAPI_Done()
{
	gapi* p = &GAPI;

	/*modify*/
	if(p->Buffer[0][0]){ Free16(p->Buffer[0][0]);p->Buffer[0][0] = NULL; }
	if(p->Buffer[1][0]){ Free16(p->Buffer[1][0]);p->Buffer[1][0] = NULL; }
	/*modify end*/

	OvlDone(&p->p);
	if (p->AygShell)
		FreeLibrary(p->AygShell);
}

#endif
