/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id: subtitle.c $
 *
 ****************************************************************************/

#include "stdafx.h"
#include <windows.h>
#include <stdlib.h>

#include "overlay_subtitle.h"
#include "../common/overlay_base.h"
#include "../common/player.h"
#include "../common/dyncode/dyncode.h"
#include "../common/blit/blit_soft.h"

#define MAXSUBTITLECOUNT		5000
#define DEFAULTFONTNAME			_T("GULIM")
#define DEFAULTHIRESFONTSIZE	28
#define DEFAULTLOWRESFONTSIZE	16
#define DEFAULTLOWRESBOUNDARY	400

//#define SUBTITLE_DEBUG

typedef struct subtitle {
	node VMT;

	int Subtitle; // Subtitle state (0:Default, 1:Not Found, 2:Load Complete)
	node *Input;
	char *String;
	tchar_t OpenedFile[MAXPATH];
	unsigned int FileSize;
	unsigned int *PositionOffset;
	unsigned int *PositionTime;

	tchar_t FontName[MAXPATH];
	int HiResFontSize;
	int LowResFontSize;
	int LowResBoundary;

	int	CurrentPos;
	int	DrawPos;

	int SrcPtr, SrcGap;

	int DstPtr, DstGap;
	int DstStepX, DstStepY;
	int DstBPP;

	int SubWidth, SubHeight;

	HFONT hFont;
	int FontSize;
	HFONT hSFont;
	int SFontSize;

	int MaxWidth;
	int MaxHeight;
	HBITMAP Bitmap;
	HGDIOBJ Bitmap0; //bitmap returned by selectobject
	HBRUSH FillBrush;
	HDC DC2;
	short *Plane;
} subtitle;

typedef struct clone_blitpack {
	blitfx FX;
	video Dst;
	video Src;
	rect DstRect;
	rect SrcRect;
	blit_soft Code[2];
} clone_blitpack;

const datadef SubtitleParams[] = 
{
	{ SUBT_ID,				TYPE_CLASS },
	{ SUBT_FONTNAME,		TYPE_STRING, DF_SETUP },
	{ SUBT_HIRESFONTSIZE,	TYPE_INT, DF_SETUP },
	{ SUBT_LOWRESFONTSIZE,	TYPE_INT, DF_SETUP },
	{ SUBT_LOWRESBOUNDARY,	TYPE_INT, DF_SETUP },

	DATADEF_END
};

void	SubtitleLoad		(subtitle* p);
void	ParseSMI			(char * str,unsigned int * pos,unsigned int * time,int size);
int 	SubtitleCreateBitmap(subtitle* p);
int		FontCreate			(subtitle* p);

#ifdef SUBTITLE_DEBUG
void PRINT(char *str){
	char tmp[128];
	FILE *fp;
	SYSTEMTIME time;
	
	fp = fopen("\\BetaLog.log", "a+");
	GetLocalTime(&time);
	sprintf(tmp, "[%02d:%02d:%02d] ", (int)time.wHour, (int)time.wMinute, (int)time.wSecond);
	fprintf(fp, tmp);
	fprintf(fp, str);
	fclose(fp);
}
#else
	#define PRINT //
#endif

static int Enum( subtitle* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,SubtitleParams,NULL);
}

static int Get( subtitle* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(SUBT_ID,int); break;
	case SUBT_FONTNAME: GETSTRING(p->FontName); break;
	case SUBT_HIRESFONTSIZE: GETVALUE(p->HiResFontSize,int); break;
	case SUBT_LOWRESFONTSIZE: GETVALUE(p->LowResFontSize,int); break;
	case SUBT_LOWRESBOUNDARY: GETVALUE(p->LowResBoundary,int); break;
	}
	return Result;
}

static int Set( subtitle* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case SUBT_FONTNAME: SETSTRING(p->FontName);Result = FontCreate(p); break;
	case SUBT_HIRESFONTSIZE: SETVALUE(p->HiResFontSize,int,Result = FontCreate(p)); break;
	case SUBT_LOWRESFONTSIZE: SETVALUE(p->LowResFontSize,int,Result = FontCreate(p)); break;
	case SUBT_LOWRESBOUNDARY: SETVALUE(p->LowResBoundary,int,Result = FontCreate(p)); break;
	}
	return Result;
}

int FontCreate(subtitle* p) {
	LOGFONT LogFont;
	node *Subtitle;

	memset(&(LogFont), 0, sizeof(LogFont));

	if(Subtitle = NodeFind(SUBT_ID)){
		if((Subtitle->Get(Subtitle, SUBT_FONTNAME, &LogFont.lfFaceName, sizeof(LogFont.lfFaceName)) != ERR_NONE)||(!wcslen(LogFont.lfFaceName))){
			wcscpy(LogFont.lfFaceName, DEFAULTFONTNAME);
			Subtitle->Set(Subtitle, SUBT_FONTNAME, &LogFont.lfFaceName, sizeof(LogFont.lfFaceName));
		}
		if((Subtitle->Get(Subtitle, SUBT_HIRESFONTSIZE, &p->HiResFontSize, sizeof(int)) != ERR_NONE)||((p->HiResFontSize < 8)||(p->HiResFontSize > 72))){
			p->HiResFontSize = DEFAULTHIRESFONTSIZE;
			Subtitle->Set(Subtitle, SUBT_HIRESFONTSIZE, &p->HiResFontSize, sizeof(int));
		}
		if((Subtitle->Get(Subtitle, SUBT_LOWRESFONTSIZE, &p->LowResFontSize, sizeof(int)) != ERR_NONE)||((p->LowResFontSize < 4)||(p->LowResFontSize > 48))){
			p->LowResFontSize = DEFAULTLOWRESFONTSIZE;
			Subtitle->Set(Subtitle, SUBT_LOWRESFONTSIZE, &p->LowResFontSize, sizeof(int));
		}
		if((Subtitle->Get(Subtitle, SUBT_LOWRESBOUNDARY, &p->LowResBoundary, sizeof(int)) != ERR_NONE)||(p->LowResBoundary < 0)){
			p->LowResBoundary = DEFAULTLOWRESBOUNDARY;
			Subtitle->Set(Subtitle, SUBT_LOWRESBOUNDARY, &p->LowResBoundary, sizeof(int));
		}
	} else {
		wcscpy(LogFont.lfFaceName, DEFAULTFONTNAME);
		p->HiResFontSize = DEFAULTHIRESFONTSIZE;
		p->LowResFontSize = DEFAULTLOWRESFONTSIZE;
		p->LowResBoundary = DEFAULTLOWRESBOUNDARY;
	}

	LogFont.lfWeight = 700;
	LogFont.lfCharSet = DEFAULT_CHARSET; //ANSI_CHARSET; //OEM_CHARSET; //HANGEUL_CHARSET; //HANGUL_CHARSET //JOHAB_CHARSET; //SHIFTJIS_CHARSET;
	LogFont.lfQuality = DEFAULT_QUALITY; //DRAFT_QUALITY; //PROOF_QUALITY;
	LogFont.lfPitchAndFamily = FF_DONTCARE;

	if((!p->hFont)||(p->FontSize != p->HiResFontSize)){
		if (p->hFont) DeleteObject(p->hFont);
		LogFont.lfHeight = p->HiResFontSize;
		p->FontSize = p->HiResFontSize;
		p->hFont = CreateFontIndirect(&LogFont);
		PRINT("HiRes Font Create.\n");
	}

	if((!p->hSFont)||(p->SFontSize != p->LowResFontSize)){
		if (p->hSFont) DeleteObject(p->hSFont);
		LogFont.lfHeight = p->LowResFontSize;
		p->SFontSize = p->LowResFontSize;
		p->hSFont = CreateFontIndirect(&LogFont);
		PRINT("LowRes Font Create.\n");
	}

	return ERR_NONE;
}

// Create DIB Bitmap for Draw Subtitle
int SubtitleCreateBitmap(subtitle* p) {
	struct {
		BITMAPINFOHEADER bmiHeader;
		int RBitMask;
		int GBitMask;
		int BBitMask;
	} Info;

	if(p->Plane != NULL) {
		PRINT("DIB Bitmap is Already Created.\n");
		return 1;
	}

	memset(&Info,0,sizeof(Info));
	Info.bmiHeader.biSize	= sizeof(Info.bmiHeader);
	Info.bmiHeader.biWidth	= p->MaxWidth;
	Info.bmiHeader.biHeight	= -(p->MaxHeight);
	Info.bmiHeader.biPlanes	= 1;
	Info.bmiHeader.biBitCount = 16;
	Info.bmiHeader.biCompression = BI_BITFIELDS;
	Info.RBitMask = 0x000000FF;
	Info.GBitMask = 0x0000FF00;
	Info.BBitMask = 0x00FF0000;

	p->Bitmap = CreateDIBSection(NULL, (BITMAPINFO*)&Info, DIB_RGB_COLORS, &p->Plane, NULL, 0);

	if (p->Bitmap) {
		if(p->Plane){
			p->Bitmap0 = SelectObject(p->DC2, p->Bitmap);
			p->FillBrush = CreateSolidBrush(RGB(0,0,0));
		} else {
			DeleteObject(p->Bitmap);
			p->Bitmap = NULL;
		}
	}

	if(p->Plane != NULL){
		PRINT("DIB Bitmap Create.\n");
	} else {
		PRINT("DIB Bitmap Create Failed.\n");
	}

	return p->Plane != NULL;
}

void ParseSMI(char * str,unsigned int * pos,unsigned int * time,int size) {
	int	Count = 1;
	char temp[512];
	char parm1[128];
	char parm2[128];
	_int64	 tpos =0;
	_int64 LastReadTime = 10;
	char *part;
	int sz=0;
	int j=0;
	int dest;

	int i;
	int LastSync	=0;
	int StartHeader	=0;

	pos[0]=0;
	time[0]=0;
	str[0]=0;

	for	(i=0;i<size;i++) {
		if	(str[i]<0) {
			i++;
			continue;
		}
		if	(str[i]>0 && str[i]<32) continue;
		if	(str[i]=='<') {
			StartHeader=i;
			continue;
		}

		if	(str[i]=='>') {
			if ((str[StartHeader+1]=='S'||str[StartHeader+1]=='s') && (str[StartHeader+2]=='Y'||str[StartHeader+2]=='y'))// SYNC TAG
			{
				// var i is closed position of SYNC Tag
				str[i]=0;str[StartHeader]=0;// Change '<', '>' to NULL
				part = &str[StartHeader+1]; // var 'part' in before '<'
				sz = strlen(part);
				for	(j=0;j<sz;j++) if(part[j]=='=') part[j]=' ';
				sscanf(part,"%s %s %d",parm1,parm2,&tpos);// get tpos(Sync timing)

				if (LastSync!=0) {
					int open = 0;
					int starttag = 0;
					int rubyopen = 0;
					part = &str[LastSync];
					strcpy(temp,part);
					// Original: Cut ascii code 0x00 ~ 0x20 character and '&nbsp;', '<br>' tag.
					//strncpy(part,temp,sz);part[sz]=0;
					//sprintf(part,"%d-%d-",Count,LastReadTime);
					//sz = strlen(part);
					//dest=dest+sz;
					dest=0;
					sz = strlen(temp);
					for (j=0;j<sz;j++) {
						if (temp[j]>0 && temp[j]<32)
							continue;
						if (temp[j]=='>') {
							int taglen = j - starttag;
							//(temp[j-1]=='r'||temp[j-1]=='R')&&(temp[j-2]=='b'||temp[j-2]=='B')
							if ((taglen == 2 || taglen == 3) && _strnicmp(temp + starttag, "br/", taglen) == 0) {
								while((dest>0)&&(part[dest-1] == ' ')){
									dest--;
								}
								part[dest]=13;dest++;
								part[dest]=10;dest++;
							}
							if (taglen == 4 && _strnicmp(temp + starttag, "ruby", 4) == 0) {
								rubyopen = 1;
							}
							if (taglen == 2 && _strnicmp(temp + starttag, "rt", 2) == 0 && rubyopen == 1) {
								rubyopen = 2;
							}
							if (taglen == 5 && _strnicmp(temp + starttag, "/ruby", 5) == 0) {
								rubyopen = 0;
							}
							open=0;
							continue;
						}
						if (temp[j]=='<') {
							open = 1;
							starttag = j+1;
							continue;
						}
						if (open == 1)
							continue;
						if (rubyopen == 2)
							continue;

						// sz>=5 && temp[j]=='&' && (temp[j+1]=='n'||temp[j+1]=='N') && (temp[j+5]==';')) {
						if (_strnicmp(temp + j, "&nbsp", 5) == 0) {
							part[dest] = ' ';
							dest++;
							if(temp[j+5] == ';')
								j++;
							j += 4;
							continue;
						}
						if (temp[j] == '&') {
							part[dest]=temp[j];
							dest++;
						}
						if ((temp[j] == ' ')&&(dest !=0)&&(part[dest-1] == ' '))
							continue;
						//if (IsDBCSLeadByte(temp[j])) { /* DoubleByte character processing */
						if (temp[j]<0) { /*fixme : processing Korean code(EUC-KR) 2byte area(0x80 ~ 0xFF)*/
							part[dest]=temp[j];
							part[dest+1]=temp[j+1];
							j++;dest=dest+2;
							continue;
						}

						part[dest]=temp[j];
						dest++;
					}
					while ((dest>0)&&(part[dest-1] == ' ')) dest--;
					part[dest]=0;
					time[Count] = (unsigned int) LastReadTime;
					pos[Count] = LastSync;
					Count++;
					if (Count >= MAXSUBTITLECOUNT) break;
					//fprintf(wfp,"%s\n",part);
				}
				LastReadTime = tpos*TICKSPERSEC/1000;// pos current
				LastSync=i+1;
			}
			continue;
		}
	}

	if (Count < MAXSUBTITLECOUNT) {
		time[Count] = 0;
		pos[Count] = size;
		Count++;
	}

	PRINT("Parse SMI.\n");
}

// Blit subtitle in screen
void BlitSubtitle(subtitle *p, short* dest) {
	int x,y;
	short *sour = p->Plane;

	if (p->Subtitle != 2 || p->SubHeight == 0) return;
	if (dest == NULL || sour == NULL){
		PRINT("Source or Dest is NULL\n");
		return; 
	}

	switch(p->DstBPP){
		case 8:{ // for ATI 3200(Y Plane)
			char *destb = (char*)dest + p->DstPtr;
			char *shade1 = (char*)destb + p->DstStepX;
			char *shade2 = (char*)destb + p->DstStepX + p->DstStepY;
			short *sourb = sour + p->SrcPtr;
			for (y=0;y<p->SubHeight;y++) {
				for (x=0;x<p->SubWidth;x++, sourb++, destb+=p->DstStepX, shade1+=p->DstStepX, shade2+=p->DstStepX) {
					if (*sourb != 0) {
						//*destb = (char)(*sourb >> 8);
						*destb = -1;
						*shade1 = 0;
						*shade2 = 0;
					}
				}
				sourb += p->SrcGap;
				destb += p->DstGap;
				shade1 += p->DstGap;
				shade2 += p->DstGap;
			}
		}
		break;
		case 16:{ // for GDI, GAPI(RGB Plane)
			short *destb = (short*)dest + p->DstPtr;
			short *shade1 = (short*)destb + p->DstStepX;
			short *shade2 = (short*)destb + p->DstStepX + p->DstStepY;
			short *sourb = sour + p->SrcPtr;
			for (y=0;y<p->SubHeight;y++) {
				for (x=0;x<p->SubWidth;x++, sourb++, destb+=p->DstStepX, shade1+=p->DstStepX, shade2+=p->DstStepX) {
					if (*sourb != 0) {
						*destb = *sourb;
						*shade1 = 0;
						*shade2 = 0;
					}
				}
				sourb += p->SrcGap;
				destb += p->DstGap;
				shade1 += p->DstGap;
				shade2 += p->DstGap;
			}
		}
		break;
	}
}

int GetSubtitlePos(subtitle *p, unsigned int curtime) {
	unsigned int time,prevtime;
	int pos = p->CurrentPos;
	int k;
	if (curtime < 1000) {
		p->Subtitle = 0;
		p->SubHeight = 0;
		p->CurrentPos = 0;
		return 0;
	}
	if (p->Subtitle==0 && curtime>1000) SubtitleLoad(p);
	if (pos<0 || pos>=MAXSUBTITLECOUNT) pos = 0;
	if (p->PositionTime == NULL) {	
		p->CurrentPos = pos;
		return 0;
	}
	for (k=0;k<100;k++) {
		if (pos<0 || pos>=(MAXSUBTITLECOUNT-1)) pos = 0;	
		time	= p->PositionTime[pos+1];
		prevtime= p->PositionTime[pos];
		if ((time == 0)&&(p->PositionOffset[pos] == p->FileSize))
			break;
		if (curtime > time)
			pos++;
		else if (curtime < prevtime)
			pos--;
		else
			break;
	}
	if(k >= 100) {
		p->DrawPos = pos;
		p->SubHeight = 0;
	}
	if (pos<0 || pos>=MAXSUBTITLECOUNT) pos = 0;
	p->CurrentPos=pos;

	return pos;
}

int DrawSubtitle(subtitle* p, overlay_base *over, int type, int pitch) {
	if (over->LastTime < 1000) {
		p->Subtitle = 0;
		p->SubHeight = 0;
		p->CurrentPos = 0;
		return -1;
	}
	if (p->Subtitle == 0 && over->LastTime>1000) SubtitleLoad(p);
	if (p->Subtitle != 2 || p->hFont == NULL || p->hSFont == NULL || p->String==NULL || p->DC2 == NULL) return -1;
	if (p->Plane == NULL && !SubtitleCreateBitmap(p)) return -1;
	if (p->CurrentPos == p->DrawPos)
		if(p->SubHeight)
			return 0;
		else
			return -1;

	{
		int pos;
		unsigned int offset;
		int destlen;
		int sourlen;
		char *sour;
		int giSourceCodePage = GetOEMCP();

		wchar_t cont1[1024];

		pos = p->CurrentPos;
		if	(pos<0 || pos>=MAXSUBTITLECOUNT) pos=0;

		offset = p->PositionOffset[pos];
		if (offset >= p->FileSize) offset = 0;

		p->CurrentPos = pos;
		sour = p->String + offset;
		sourlen = strlen(sour);

		if (sourlen>=1000) {
			sour[999]=0;
			sour[998]=0;
			sourlen = strlen(sour);
		}

		destlen = MultiByteToWideChar(giSourceCodePage, MB_PRECOMPOSED , sour , sourlen, NULL, 0 );

		if	(destlen < 1000)
			destlen = MultiByteToWideChar(giSourceCodePage, 0, sour, sourlen, (LPWSTR)cont1, destlen);
		else
			destlen = 0;

		if(destlen > 0){
			int DstWidth, DstHeight;
			int DstX, DstY;
			int DstPitch;
			int bottomgap;
			COLORREF crColor[2] = {0x00FFFFFF,0x00000000};
			HFONT obj;
			RECT tRect;
			clone_blitpack *Pack = (clone_blitpack*)over->Soft;

			cont1[destlen] = 0;

			if(type & SUBTITLE_GAPI){
				DstWidth = over->DstAlignedRect.Width;
				DstHeight = over->DstAlignedRect.Height;
				DstX = over->DstAlignedRect.x;
				DstY = over->DstAlignedRect.y;
			} else if(type & SUBTITLE_INTEL2700G){
				DstWidth = Pack->Dst.Width;
				DstHeight = Pack->Dst.Height;
				DstX = 0;
				DstY = 0;
			} else {
				DstWidth = over->DstAlignedRect.Width;
				DstHeight = over->DstAlignedRect.Height;
				DstX = 0;
				DstY = 0;
			}
			if(Pack->Code[0].SwapXY)
				SwapInt(&DstWidth, &DstHeight);

			p->SubWidth = DstWidth;
			if(p->SubWidth > p->MaxWidth)
				p->SubWidth = p->MaxWidth;

			bottomgap = 4;
			if(p->SubWidth < p->LowResBoundary) bottomgap >>= 1;

			tRect.left = 0;
			tRect.top = 0;
			tRect.right = p->SubWidth;
			tRect.bottom = DstHeight - bottomgap;

			if(tRect.right < p->LowResBoundary) {
				tRect.bottom = (int)(tRect.bottom / (p->LowResFontSize)) * (p->LowResFontSize);
				obj = SelectObject(p->DC2, p->hSFont);
			} else {
				tRect.bottom = (int)(tRect.bottom / p->HiResFontSize) * (p->HiResFontSize);
				obj = SelectObject(p->DC2, p->hFont);
			}

			SetBkColor(p->DC2, crColor[1]);
			SetTextColor(p->DC2, crColor[0]);

			FillRect(p->DC2,&tRect, p->FillBrush);

			p->SubHeight = DrawText(p->DC2, cont1, -1, &tRect, DT_WORDBREAK | DT_CENTER);

			if (obj != NULL) SelectObject(p->DC2, obj);

			if(p->SubHeight > 0){
				p->DstBPP = Pack->Code[0].DstBPP;
				p->SrcGap = p->MaxWidth - p->SubWidth;
				p->SrcPtr = 0;

				if(type & (SUBTITLE_ATI3200)){
					if(pitch == -1)
						DstPitch = over->DstAlignedRect.Width;
					else
						DstPitch = pitch;
					if(DstPitch & 15) DstPitch += 16 - (DstPitch & 15);
					if(!Pack->Code[0].SwapXY){
						p->DstGap = DstPitch - p->SubWidth;
						p->DstPtr = (DstY + DstHeight - p->SubHeight) * DstPitch + DstX;
						p->DstStepX = 1;
						p->DstStepY = DstPitch;
					} else {
						p->DstGap = -(p->SubWidth * DstPitch) + 1;
						p->DstPtr = DstY * DstPitch + DstX + DstHeight - p->SubHeight;
						p->DstStepX = DstPitch;
						p->DstStepY = 1;
					}
				} else  if(type & (SUBTITLE_GDI | SUBTITLE_GAPI | SUBTITLE_INTEL2700G)){
					if(pitch == -1)
						DstPitch = Pack->Dst.Pitch / (p->DstBPP >> 3);
					else
						DstPitch = pitch;

					p->DstStepX = 1;
					p->DstStepY = DstPitch;

					if (over->FX.Direction & DIR_MIRRORLEFTRIGHT)
						p->DstStepX = -p->DstStepX;

					if (over->FX.Direction & DIR_MIRRORUPDOWN)
						p->DstStepY = -p->DstStepY;

					if (over->FX.Direction & DIR_SWAPXY)
						SwapInt(&p->DstStepX, &p->DstStepY);

					switch(over->FX.Direction & (DIR_SWAPXY | DIR_MIRRORLEFTRIGHT | DIR_MIRRORUPDOWN)){
					case 0:
						p->DstGap = DstPitch - p->SubWidth;
						p->DstPtr = (DstY + DstHeight - p->SubHeight) * DstPitch + DstX;
						break;
					case DIR_MIRRORLEFTRIGHT:
						p->DstGap = DstPitch + p->SubWidth;
						p->DstPtr = (DstY + DstHeight - p->SubHeight) * DstPitch + DstX + DstWidth;
						break;
					case DIR_MIRRORUPDOWN:
						p->DstGap = -(DstPitch + p->SubWidth);
						p->DstPtr = (DstY + p->SubHeight) * DstPitch + DstX;
						break;
					case DIR_MIRRORLEFTRIGHT | DIR_MIRRORUPDOWN:
						p->DstGap = -(DstPitch - p->SubWidth);
						p->DstPtr = (DstY + p->SubHeight + bottomgap) * DstPitch + DstX + DstWidth;
						break;
					case DIR_SWAPXY:
						p->DstGap = (p->SubWidth * DstPitch) - 1;
						p->DstPtr = (DstY + p->SubWidth) * DstPitch + DstX + p->SubHeight + bottomgap;
						break;
					case DIR_SWAPXY | DIR_MIRRORLEFTRIGHT:
						p->DstGap = -(p->SubWidth * DstPitch) - 1;
						p->DstPtr = DstY * DstPitch + DstX + p->SubHeight + bottomgap;
						break;
					case DIR_SWAPXY | DIR_MIRRORUPDOWN:
						p->DstGap = (p->SubWidth * DstPitch) + 1;
						p->DstPtr = (DstY + DstWidth) * DstPitch + DstX + DstHeight - p->SubHeight;
						break;
					case DIR_SWAPXY | DIR_MIRRORLEFTRIGHT | DIR_MIRRORUPDOWN:
						p->DstGap = -(p->SubWidth * DstPitch) + 1;
						p->DstPtr = DstY * DstPitch + DstX + DstHeight - p->SubHeight;
						break;
					}
				}

			#ifdef SUBTITLE_DEBUG
				{
					char temp[256];

					sprintf(temp,
						"1: bpp:%d dw:%d dh:%d dx:%d dy:%d sx:%d sy:%d dw:%d dh:%d drx:%d dry:%d drw:%d drh:%d sp:%d dp:%d sg:%d dg:%d dn:%d ln:%d st:%d pos:%d\n",
						p->DstBPP,
						DstWidth, DstHeight, DstX, DstY,
						p->SubWidth, p->SubHeight,
						Pack->Dst.Width, Pack->Dst.Height,
						Pack->DstRect.x, Pack->DstRect.y,
						Pack->DstRect.Width, Pack->DstRect.Height,
						p->SrcPtr, p->DstPtr,
						p->SrcGap, p->DstGap, p->DstStepX, p->DstStepY,
						p->Subtitle, p->CurrentPos);
					PRINT(temp);

					sprintf(temp,
						"2: dp:%d dsx:%d dsy:%d drx:%d dry:%d dp:%d dg:%d lr:%d up:%d xy:%d\n",
						DstPitch,
						p->DstStepX,
						p->DstStepY,
						DstX,
						DstY,
						p->DstPtr,
						p->DstGap,
						Pack->Code[0].DstLeftRight,
						Pack->Code[0].DstUpDown,
						Pack->Code[0].SwapXY);
					PRINT(temp);
				}
			#endif
			}
		} else {
			p->SubHeight = 0;
		}

		p->DrawPos = pos;
	}
	return 1;
}

void SubtitleLoad(subtitle *p) {	
	FILE *fp;
	int	len;
	tchar_t s1[MAXPATH];
	tchar_t s2[MAXPATH];
	datadef DataDef; 

	PRINT("Subtitle Load.\n");

	if (p->DC2 == NULL) {
		PRINT("Memory DC is NULL(Terminate).\n");
		p->OpenedFile[0] = 0;
		p->Subtitle = 1;
		return;
	}
#ifdef SUBTITLE_DEBUG
	if (p->Plane == NULL) {
		PRINT("Bitmap Plane is Destoryed.\n");
		return;
	}
#endif
	if (p->Plane == NULL && !SubtitleCreateBitmap(p)) {
		PRINT("DIB Bitmap Create Failed.\n");
		p->OpenedFile[0] = 0;
		p->Subtitle = 1;
		return;
	}

	if(p->Input == NULL){
		node *Player = NodeFind(PLAYER_ID);
		if(Player == NULL){
			PRINT("Get Player Node Failed.\n");
			return;
		}

		if((Player->Get(Player, PLAYER_INPUT, &p->Input, sizeof(p->Input))!=ERR_NONE)||(p->Input == NULL)){
			PRINT("Get Input Node Failed.\n");
			return;
		}
	}

	p->Input->Enum(p->Input, 4, &DataDef);
	p->Input->Get(p->Input, DataDef.No, s2, sizeof(s2));
	// s2 is playing file name. (Change extension to '.smi')
	if(wcsicmp(s2, p->OpenedFile) != 0){
		// Release preloaded subtitle.
		if(p->String) free(p->String);
		p->String = NULL;
		p->FileSize = 0;
		p->CurrentPos = 0;

		wcscpy(p->OpenedFile, s2);

		//PRINT(DataDef.Name);PRINT(" , ");PRINT(s2);
		wcscpy(s1, s2); // copy s2 to s1
		len=wcslen(s2);
		s1[len-3]=0; // cut s1's extension part
		wsprintf(s2,_T("%sSMI"),s1); // concatenate extension 'smi'
		//PRINT(" -> ");PRINT(s2);PRINT("\n");

		fp = _wfopen(s2, _T("rb"));

		if (fp == NULL) {
			p->OpenedFile[0] = 0;
			p->Subtitle = 1;
			PRINT("Subtitle File not found.\n");
		} else {
			fseek(fp,0,SEEK_END);
			p->FileSize = ftell(fp);
			fseek(fp,0,SEEK_SET);

			p->String = malloc(p->FileSize+32); // 32 byte is buffer.
			if (p->String) {
				fread(p->String, p->FileSize, 1, fp);
				fclose(fp);
				ParseSMI(p->String,p->PositionOffset,p->PositionTime,p->FileSize);
				p->Subtitle = 2;
				PRINT("Subtitle Load Success.\n");
			} else {
				p->OpenedFile[0] = 0;
				p->Subtitle = 0;
				PRINT("Subtitle Load Failed.\n");
			}
		}
	} else {
		PRINT("Subtitle Already Loaded.\n");
		p->Subtitle = 2;
	}

	RedrawSubtitle(p);
}

void RedrawSubtitle(subtitle *p){
	PRINT("Redraw Subtitle.\n");
	p->DrawPos = -1;
	p->SubHeight = 0;
}

static subtitle SubTitle;

void SubTitle_Init()
{
	subtitle* p = &SubTitle;
	NodeFill(&p->VMT,sizeof(subtitle),Enum,Get,Set);
	tcscpy(p->FontName, DEFAULTFONTNAME);
	p->HiResFontSize = DEFAULTHIRESFONTSIZE;
	p->LowResFontSize = DEFAULTLOWRESFONTSIZE;
	p->LowResBoundary = DEFAULTLOWRESBOUNDARY;

	p->OpenedFile[0] = 0;
	p->Subtitle = 0;
	p->FileSize = 0;
	p->String = NULL;
	p->Input = NULL;

	p->Plane = NULL;
	p->Bitmap = NULL;
	p->PositionOffset = malloc(MAXSUBTITLECOUNT * sizeof(int));
	p->PositionTime = malloc(MAXSUBTITLECOUNT * sizeof(int));

	p->MaxWidth = GetSystemMetrics(SM_CXSCREEN);
	p->MaxHeight = GetSystemMetrics(SM_CYSCREEN);
	if(p->MaxWidth > p->MaxHeight)
		p->MaxHeight = p->MaxWidth;
	else
		p->MaxWidth = p->MaxHeight;

	p->DC2 = CreateCompatibleDC(NULL);
	p->hFont = NULL;
	p->hSFont = NULL;
	p->FontSize = 0;
	p->SFontSize = 0;

	NodeRegister(&p->VMT,PRI_DEFAULT);

	SubtitleCreateBitmap(p);
	FontCreate(p);
}

void SubTitle_Done()
{
	subtitle* p = &SubTitle;

	if (p->Bitmap) {
		SelectObject(p->DC2, p->Bitmap0);
		DeleteObject(p->Bitmap);
		p->Bitmap = NULL;

		DeleteObject(p->FillBrush);
		p->FillBrush = NULL;
	}
	p->Plane = NULL;
	DeleteDC(p->DC2);
	p->DC2 = NULL;

	if (p->String) free(p->String);
	p->String = NULL;
	if (p->PositionOffset) free(p->PositionOffset);
	p->PositionOffset=NULL;
	if (p->PositionTime) free(p->PositionTime);
	p->PositionTime=NULL;
	p->OpenedFile[0] = 0;

	if (p->hFont) DeleteObject(p->hFont);
	p->hFont = NULL;
	if (p->hSFont) DeleteObject(p->hSFont);
	p->hSFont = NULL;

	p->Subtitle = 0;
	p->SubHeight = 0;

	NodeUnRegister(&p->VMT);

	PRINT("Subtitle Done\n\n");
}
